@@ -17,8 +17,10 @@
 import collections
 from dataclasses import dataclass
 import functools
+import warnings
 
 
+from flax import traceback_util
 import jax
 from jax import random
 
@@ -28,10 +30,11 @@ from .frozen_dict import freeze
 from .frozen_dict import FrozenDict
 from .frozen_dict import unfreeze
 
-from .scope import Scope, CollectionFilter, PRNGSequenceFilter, in_filter, union_filters, intersect_filters, subtract_filters, group_collections
+from .scope import Scope, DenyList, CollectionFilter, PRNGSequenceFilter, in_filter, union_filters, intersect_filters, subtract_filters, group_collections
 
 from . import axes_scan
 
+traceback_util.register_exclusion(__file__)
 
 T = TypeVar('T')
 
@@ -74,13 +77,17 @@ def pack(fn: Callable[..., Any],
          in_variable_filters: Sequence[CollectionFilter],
          out_variable_filters: Sequence[CollectionFilter],
          rng_filters: Sequence[PRNGSequenceFilter],
-         name=None) -> Callable[..., Any]:
+         name=None,
+         enable_kwargs=False) -> Callable[..., Any]:
   """Pack variables and rngs for functional transformations.
 
   The pack function is the building block for all other lifted transformations.
   """
   @functools.wraps(fn)
-  def wrapper(scope_tree: Scope, *args):
+  def wrapper(scope_tree: Scope, *args, **kwargs):
+    if not enable_kwargs and kwargs:
+      msg = 'kwargs are not supported in {}, so \"{}\" is(are) ignored'
+      warnings.warn(msg.format(name, ', '.join(kwargs.keys())), RuntimeWarning)
     # pylint: disable=protected-access
     scopes, treedef = jax.tree_flatten(scope_tree)
     scopes, paths = _dedup_scopes(scopes)
@@ -104,12 +111,16 @@ def pack(fn: Callable[..., Any],
           if not col_in_out:
             variable_group[col_name] = freeze(collection)
     rng_groups_xs = []
+    inner_rng_counters = []
     for scope in scopes:
+      rng_counters = {}
       rng_groups = group_collections(scope.rngs, rng_filters)
       for rng_group in rng_groups:
         for kind in rng_group:
-          rng_group[kind] = scope.make_rng(kind)
+          rng_group[kind] = scope.rngs[kind]
+          rng_counters[kind] = scope.rng_counters[kind]
       rng_groups_xs.append(rng_groups)
+      inner_rng_counters.append(rng_counters)
     rng_groups_xs_t = _transpose(rng_groups_xs)
 
     inner_scopes = []
@@ -128,7 +139,7 @@ def pack(fn: Callable[..., Any],
       rng_groups_xs = _transpose(rng_groups_xs_t) or ((),) * len(scopes)
       assert len(variable_groups_xs) == len(scopes)
       assert len(rng_groups_xs) == len(scopes)
-      for variable_groups, rng_groups, scope in zip(variable_groups_xs, rng_groups_xs, scopes):
+      for variable_groups, rng_groups, scope, rng_counters in zip(variable_groups_xs, rng_groups_xs, scopes, inner_rng_counters):
         variables = {}
         rngs = {}
         for variable_group in variable_groups:
@@ -148,6 +159,7 @@ def pack(fn: Callable[..., Any],
             variables, name=scope.name, rngs=rngs,
             mutable=scope_mutable, parent=None,
             path=new_path)
+        inner_scope.rng_counters = rng_counters
         inner_scopes.append(inner_scope)
       inner_scopes = _dup_scopes(scopes, inner_scopes, paths)
       return treedef.unflatten(inner_scopes)
@@ -174,19 +186,28 @@ def pack(fn: Callable[..., Any],
       return _transpose(out_variable_groups_xs)
 
     try:
-      y, out_variable_groups_xs_t = fn(
-          scope_fn, repack,
-          variable_groups_xs_t, rng_groups_xs_t,
-          *args)
+      if enable_kwargs:
+        y, out_variable_groups_xs_t = fn(
+            scope_fn, repack,
+            variable_groups_xs_t, rng_groups_xs_t,
+            *args, **kwargs)
+      else:
+        y, out_variable_groups_xs_t = fn(
+            scope_fn, repack,
+            variable_groups_xs_t, rng_groups_xs_t,
+            *args)
     finally:
       for inner_scope in inner_scopes:
         inner_scope.invalidate()
     out_variable_groups_xs = _transpose(out_variable_groups_xs_t)
-    for scope, out_variable_groups in zip(scopes, out_variable_groups_xs):
+    for scope, out_variable_groups, rng_counters in zip(scopes, out_variable_groups_xs, inner_rng_counters):
       for out_variable_group in out_variable_groups:
         for col_name, collection in out_variable_group.items():
           for var_name, value in collection.items():
             scope.put_variable(col_name, var_name, value)
+      for kind, rng_counter in rng_counters.items():
+        assert rng_counter >= scope.rng_counters[kind]
+        scope.rng_counters[kind] = rng_counter
     return y
   return wrapper
 
@@ -210,7 +231,7 @@ def transform(
     trans_out_fn: transforms the updated variables in the view after mutation.
     init: If True, variables are initialized before transformation.
     rngs: PRNGSequences added to the transformed scope (default: all).
-    variables: Addtional Variable collections added to the transformed scope.
+    variables: Additional Variable collections added to the transformed scope.
       Besides those specified by `target` (default: all).
   """
   def wrapper(scope_fn, repack, variable_groups, rng_groups, treedef, *args):
@@ -420,7 +441,9 @@ def scan(fn: Callable[..., Any],
          split_rngs: Mapping[PRNGSequenceFilter, bool] = {},
          in_axes=0, out_axes=0,
          length: Optional[int] = None,
-         reverse: bool = False) -> Callable[..., Any]:
+         reverse: bool = False,
+         data_transform: Optional[Callable[..., Any]] = None,
+         ) -> Callable[..., Any]:
   """A lifted version of ``jax.lax.scan``.
 
   See ``jax.lax.scan`` for the unlifted scan in Jax.
@@ -478,6 +501,9 @@ def scan(fn: Callable[..., Any],
     length: Specifies the number of loop iterations. This only needs
       to be specified if it cannot be derivied from the scan arguments.
     reverse: If true, scan from end to start in reverse order.
+    data_transform: optional function to transform raw variable and rng groups,
+      intended for inline SPMD annotations.
+
   Returns:
     The scan function with the signature ``(scope, carry, *xxs) -> (carry, yys)``,
     where ``xxs`` and ``yys`` are the scan values that go in and out of the loop.
@@ -524,6 +550,9 @@ def scan(fn: Callable[..., Any],
     def scanned(broadcast_vars, carry, scan_variable_groups, rng_groups, args):
       carry_vars, c = carry
       variable_groups = (broadcast_vars, carry_vars) + scan_variable_groups
+      if data_transform is not None:
+        variable_groups, rng_groups = data_transform(variable_groups,
+                                                     rng_groups)
       scope = scope_fn(variable_groups, rng_groups)
       c, y = fn(scope, c, *args)
       out_vars = repack_fn(scope)
@@ -584,10 +613,10 @@ def custom_vjp(fn: Callable[..., Any], backward_fn: Callable[..., Any],
     dense_sign_grad = lift.custom_vjp(fwd, backward_fn=bwd, nondiff_argnums=(2,))
 
   Args:
-    fn: should return a tuple of output and auxilliary data for the backward pass.
+    fn: should return a tuple of output and auxiliary data for the backward pass.
     backward_fn: arguments are passed as (*nondiff_args, scope_fn, grad_variables, aux, g_y)
       where scope_fn takes grad_variables to create the scope,
-      aux is the auxilliary data returend by `fn`,
+      aux is the auxiliary data returned by `fn`,
       and g_y is the tangent of y.
   """
   # TODO(jheek) is this transform general/flexible enough?
@@ -634,6 +663,7 @@ def checkpoint(fn: Callable[..., Any],
                variables: CollectionFilter = True,
                rngs: PRNGSequenceFilter = True,
                concrete: bool = False,
+               prevent_cse: bool = True,
                ) -> Callable[..., Any]:
   """Lifted version of ``jax.checkpoint``.
 
@@ -644,29 +674,48 @@ def checkpoint(fn: Callable[..., Any],
     re-computed when computing gradients.
     variables: The variable collections that are lifted. By default all
       collections are lifted.
-    rngs: The PRNG sequences that are lifted. By defualt all PRNG sequences
+    rngs: The PRNG sequences that are lifted. By default all PRNG sequences
       are lifted.
     concrete: Optional, boolean indicating whether ``fun`` may involve
       value-dependent Python control flow (default False). Support for such
       control flow is optional, and disabled by default, because in some
       edge-case compositions with :func:`jax.jit` it can lead to some extra
       computation.
+    prevent_cse: Optional, boolean indicating whether to prevent common
+      subexpression elimination (CSE) optimizations in the HLO generated from
+      differentiation. This CSE prevention has costs because it can foil other
+      optimizations, and because it can incur high overheads on some backends,
+      especially GPU. The default is True because otherwise, under a ``jit`` or
+      ``pmap``, CSE can defeat the purpose of this decorator. But in some
+      settings, like when used inside a ``scan``, this CSE prevention mechanism
+      is unnecessary, in which case ``prevent_cse`` can be set to False.
   Returns:
     A wrapped version of ``fn``. When computing gradients intermediate
     computations will be re-computed when computing gradients.
   """
-  def inner(scope_fn, repack_fn, variable_groups, rng_groups, *args):
-    @functools.partial(jax.remat, concrete=concrete)
+  def inner(scope_fn, repack_fn, variable_groups, rng_groups, *args, **kwargs):
+    @functools.partial(jax.remat, concrete=concrete, prevent_cse=prevent_cse)
     @functools.wraps(fn)
-    def rematted(variable_groups, rng_groups, *args):
+    def rematted(variable_groups, rng_groups, *args, **kwargs):
       scope = scope_fn(variable_groups, rng_groups)
-      y = fn(scope, *args)
+      y = fn(scope, *args, **kwargs)
       return y, repack_fn(scope)
 
-    return rematted(variable_groups, rng_groups, *args)
-  return pack(inner, (variables,), (variables,), (rngs,), name='remat')
+    return rematted(variable_groups, rng_groups, *args, **kwargs)
+  return pack(inner, (variables,), (variables,), (rngs,), name='remat', enable_kwargs=True)
 
 
+remat = checkpoint
+
+
+def _hashable_filter(x):
+  """Hashable version of CollectionFilter."""
+  if isinstance(x, Iterable):
+    return tuple(x)  # convert un-hashable list & sets to tuple
+  if isinstance(x, DenyList):
+    return DenyList(_hashable_filter(x.deny))  # convert inner filter recursively
+  return x
+
 def jit(fn: Callable[..., Any],
         variables: CollectionFilter = True,
         rngs: PRNGSequenceFilter = True,
@@ -676,12 +725,12 @@ def jit(fn: Callable[..., Any],
         backend: Union[str, None] = None,
         ) -> Callable[..., Any]:
   """Lifted version of ``jax.jit``.
-  
+
   Args:
     fn: Scope function to be jitted.
     variables: The variable collections that are lifted. By default all
       collections are lifted.
-    rngs: The PRNG sequences that are lifted. By defualt all PRNG sequences
+    rngs: The PRNG sequences that are lifted. By default all PRNG sequences
       are lifted.
     static_argnums: An int or collection of ints specifying which positional
       arguments to treat as static (compile-time constant). Operations that only
@@ -715,28 +764,44 @@ def jit(fn: Callable[..., Any],
     static_argnums = (static_argnums,)
   if not isinstance(donate_argnums, Iterable):
     donate_argnums = (donate_argnums,)
-  # offset argnums by one because first argument is the scope.
-  static_argnums = tuple(i + 1 for i in static_argnums if i > 0)
-  donate_argnums = tuple(i + 1 for i in donate_argnums if i > 0)
-  def inner(scope_fn, repack_fn, variable_groups, rng_groups, *args):
-    @functools.partial(jax.jit,
-                       static_argnums=static_argnums,
-                       donate_argnums=donate_argnums,
-                       device=device, backend=backend)
-    @functools.wraps(fn)
-    def jitted(variable_groups, rng_groups, *args):
-      scope = scope_fn(variable_groups, rng_groups)
-      y = fn(scope, *args)
-      return y, repack_fn(scope)
+  # offset argnums by two because first argument in the original function is the scope
+  # while jitted has 3 functions before the user arguments.
+  static_argnums = (0,) + tuple(i + 2 for i in static_argnums if i > 0)
+  donate_argnums = tuple(i + 2 for i in donate_argnums if i > 0)
 
-    return jitted(variable_groups, rng_groups, *args)
+  # Close over scope_fn & repack_fn to avoid recompilation
+  # this is impure but we use the fingerprint arg to differentiate between cases
+  # where scope_fn or repack_fn actually produce non-identical results.
+  scope_fn = None  # type: Callable
+  repack_fn = None  # type: Callable
+  @functools.partial(jax.jit,
+                     static_argnums=static_argnums,
+                     donate_argnums=donate_argnums,
+                     device=device, backend=backend)
+  @functools.wraps(fn)
+  def jitted(fingerprint, variable_groups, rng_groups, *args):
+    nonlocal scope_fn, repack_fn
+    # fingerprint is only used to differentiate the cache signature for cases
+    # where different collections are mutable.
+    del fingerprint
+    scope = scope_fn(variable_groups, rng_groups)  # pylint: disable=not-callable
+    y = fn(scope, *args)
+    return y, repack_fn(scope)  # pylint: disable=not-callable
+
+  def inner(scope_fun, repack_fun, variable_groups, rng_groups, *args):
+    nonlocal scope_fn, repack_fn
+    try:
+      scope_fn = scope_fun
+      repack_fn = repack_fun
+      scopes = jax.tree_leaves(scope_fn(variable_groups, rng_groups))
+      mutable = tuple(_hashable_filter(scope.mutable) for scope in scopes)
+      return jitted(mutable, variable_groups, rng_groups, *args)
+    finally:
+      scope_fn, repack_fn = None, None
 
   return pack(inner, (variables,), (variables,), (rngs,), name='jit')
 
 
-remat = checkpoint
-
-
 def remat_scan(body_fn: Callable[..., Any], scope: Scope, carry: Any,
                lengths: Sequence[int],
                variable_broadcast: CollectionFilter = False,
