@@ -22,12 +22,16 @@ from typing import Any, Callable, Container, Dict, Generic, Iterable, Mapping, O
 
 from . import tracers
 from flax import errors
+from flax import traceback_util
 from .frozen_dict import freeze
 from .frozen_dict import FrozenDict
 from .frozen_dict import unfreeze
 import jax
 from jax import numpy as jnp
 from jax import random
+import numpy as np
+
+traceback_util.register_exclusion(__file__)
 
 T = TypeVar('T')
 
@@ -41,6 +45,18 @@ Filter = Union[bool, str, Container[str], 'DenyList']
 
 @dataclasses.dataclass(frozen=True, eq=True)
 class DenyList:
+  """DenyList represents an opt-out based mutability filter.
+  
+  DenyList can be used to make every collection mutable except the ones
+  defined in the given filter.
+  To for example make everything but the params collection mutable::
+
+    nn.apply(fn, mutable=nn.DenyList(["params"]))
+
+  Attributes:
+    deny: The filter representing the collections that are not mutable.
+
+  """
   deny: Filter
 
 
@@ -59,7 +75,7 @@ def _fold_in_str(rng: PRNGKey, data: str) -> PRNGKey:
   """Folds a string into a jax.random.PRNGKey using its SHA-1 hash.
 
   This is faster than splitting an PRNGKey because it allows generating new PRNG
-  keys in parellel that are independent of each other.
+  keys in parallel that are independent of each other.
 
   Args:
    rng: the rng to fold the string into.
@@ -271,7 +287,7 @@ class Scope:
   usually generally do not interact with ``Scopes`` directly.
 
   See `core design tests
-  <https://github.com/google/flax/tree/master/tests/core/design>`_
+  <https://github.com/google/flax/tree/main/tests/core/design>`_
   for a number of examples using ``Scopes``.
   """
 
@@ -281,7 +297,7 @@ class Scope:
                name: Optional[str] = None,
                mutable: CollectionFilter = False,
                parent: Optional['Scope'] = None,
-               path: Tuple[str] = ()):
+               path: Iterable[str] = ()):
     """Initializes a Scope.
 
     Args:
@@ -295,7 +311,7 @@ class Scope:
     self._variables = variables
     self.parent = parent
     self.name = name
-    self.path = path
+    self.path = tuple(path)
     self.rngs = rngs if rngs else {}
     self.mutable = mutable
 
@@ -636,7 +652,7 @@ def bind(variables: VariableDict,
   """Bind variables and rngs to a new ``Scope``.
   
   bind provides a ``Scope`` instance without transforming a function
-  with ``apply``. This is particulary useful for debugging and
+  with ``apply``. This is particalary useful for debugging and
   interactive use cases like notebooks where a function would limit
   the ability split up code into different cells.
 
@@ -734,7 +750,7 @@ def _is_valid_variables(variables: VariableDict) -> bool:
 
 
 def _is_valid_rng(rng: Array):
-  if not isinstance(rng, jnp.ndarray):
+  if not isinstance(rng, (np.ndarray, jnp.ndarray)):
     return False
   if rng.shape != (2,) or rng.dtype != jnp.uint32:
     return False
