@@ -63,7 +63,7 @@ class Template(FlaxError):
 
 class FlaxError(Exception):
   def __init__(self, message):
-    error_page = 'https://flax.readthedocs.io/en/improve-error/flax.errors.html'
+    error_page = 'https://flax.readthedocs.io/en/latest/flax.errors.html'
     module_name = self.__class__.__module__
     class_name = self.__class__.__name__
     error_msg = f'{message} ({error_page}#{module_name}.{class_name})'
@@ -166,7 +166,7 @@ class ScopeParamNotFoundError(FlaxError):
       return embedding[inputs]
 
     variables = Embed(4, 8).init(random.PRNGKey(0), jnp.ones((5, 5, 1)))
-    _ = NoBiasDense().apply(variables, jnp.ones((5, 5, 1)), 'embed')
+    _ = Embed().apply(variables, jnp.ones((5, 5, 1)), 'embed')
   """
   def __init__(self, param_name, scope_path):
     super().__init__(f'No parameter named "{param_name}" exists in '
@@ -178,7 +178,7 @@ class ScopeParamShapeError(FlaxError):
   This error is thrown when the shape of an existing parameter is different from
   the shape of the return value of the ``init_fn``. This can happen when the 
   shape provided during :meth:`Module.apply() <flax.linen.Module.apply>` is
-  different from the one used when intializing the module.
+  different from the one used when initializing the module.
   
   For instance, the following code throws this error because the apply shape 
   (``(5, 5, 1)``) is different from the init shape (``(5, 5``). As a result, the
@@ -263,6 +263,19 @@ class ModifyScopeVariableError(FlaxError):
                      f'"{scope_path}" because collection "{col}" is immutable.')
 
 
+class JaxTransformError(FlaxError):
+  """
+  JAX transforms and Flax modules cannot be mixed.
+
+  JAX's functional transformations expect pure function.
+  When you want to use JAX transformations **inside** Flax models,
+  you should make use of the Flax transformation wrappers
+  (e.g.: ``flax.linen.vmap``, ``flax.linen.scan``, etc.).
+  """
+  def __init__(self):
+    super().__init__('Jax transforms and Flax models cannot be mixed.')
+
+
 #################################################
 # module.py errors                              #
 #################################################
@@ -357,7 +370,7 @@ class AssignSubModuleError(FlaxError):
     Foo().init(random.PRNGKey(0), jnp.zeros((1,)))
 
   In this case, ``self.conv(kernel_size=4)`` is called from ``__call__``, which
-  is disallowed beause it's neither within ``setup`` nor a method wrapped in
+  is disallowed because it's neither within ``setup`` nor a method wrapped in
   x``nn.compact``.
   """
   def __init__(self, cls):
@@ -505,18 +518,6 @@ class CallCompactUnboundModuleError(FlaxError):
     super().__init__('Can\'t call compact methods on unbound modules')
 
 
-class JaxOmnistagingError(FlaxError):
-  """
-  The Flax linen API requires JAX omnistaging to be enabled. In order to enable
-  this, add this to your imports::
-    
-    from jax.config import config
-    config.enable_omnistaging()
-  """
-  def __init__(self):
-    super().__init__(f'Flax Linen requires Omnistaging to be enabled')
-
-
 class InvalidCheckpointError(FlaxError):
   """
   A checkpoint cannot be stored in a directory that already has
@@ -527,3 +528,22 @@ class InvalidCheckpointError(FlaxError):
   """
   def __init__(self, path, step):
     super().__init__(f'Trying to save an outdated checkpoint at step: "{step}" and path: "{path}".')
+
+
+#################################################
+# transforms.py errors                          #
+#################################################
+
+class TransformedMethodReturnValueError(FlaxError):
+  """
+  Transformed Module methods cannot return other Modules or Variables.
+
+  This commonly occurs when named_call is automatically applied to helper
+  constructor methods when profiling is enabled, and can be mitigated by
+  using the @nn.nowrap decorator to prevent automatic wrapping.
+  """
+  def __init__(self, name):
+    super().__init__(
+      f'Transformed module method {name} cannot return Modules or Variables. '
+      f'For helper constructor methods use the @nn.nowrap decorator to prevent '
+      f'decoration by the automatic named_call transform.')
