@@ -12,19 +12,6 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-# Copyright 2021 The Flax Authors.
-#
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
-#
-#     http://www.apache.org/licenses/LICENSE-2.0
-#
-# Unless required by applicable law or agreed to in writing, software
-# distributed under the License is distributed on an "AS IS" BASIS,
-# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-# See the License for the specific language governing permissions and
-# limitations under the License.
 """Utilities we could consider upstreaming to Jax.
 """
 
@@ -49,13 +36,13 @@ def _replicate(x, devices=None):
     # match the default device assignments used in pmap:
     # for single-host, that's the XLA default device assignment
     # for multi-host, it's the order of jax.local_devices()
-    if jax.host_count() == 1:
+    if jax.process_count() == 1:
       devices = [d for d in xb.get_backend().get_default_device_assignment(
-          jax.device_count()) if d.host_id == jax.host_id()]
+          jax.device_count()) if d.process_index == jax.process_index()]
     else:
       devices = jax.local_devices()
-  if hasattr(jax.api, "device_put_sharded"):  # jax >= 0.2.0
-    return jax.api.device_put_sharded(len(devices) * [x], devices)
+  if hasattr(jax, "device_put_sharded"):  # jax >= 0.2.0
+    return jax.device_put_sharded(len(devices) * [x], devices)
   else:
     aval = jax.ShapedArray((len(devices),) + x.shape, x.dtype)
     buffers = [xla.device_put(x, device=d) for d in devices]
@@ -113,12 +100,7 @@ def partial_eval_by_shape(fn, input_spec, *args, **kwargs):
   f_flat, out_tree = jax.api_util.flatten_fun_nokwargs(lu.wrap_init(f), in_tree)
   in_pvals = [pe.PartialVal.unknown(jax.ShapedArray(x.shape, x.dtype))
               for x in inputs_flat]
-
-  if config.omnistaging_enabled:
-    _, out_pvals, _ = pe.trace_to_jaxpr(f_flat, in_pvals)
-  else:
-    with jax.core.initial_style_staging():
-      _, out_pvals, _ = pe.trace_to_jaxpr(f_flat, in_pvals, stage_out=True)
+  _, out_pvals, _ = pe.trace_to_jaxpr(f_flat, in_pvals)
   out_flat = [const if pv is None else jax.ShapeDtypeStruct(pv.shape, pv.dtype)
               for pv, const in out_pvals]
   return jax.tree_unflatten(out_tree(), out_flat)
@@ -155,8 +137,8 @@ def prefetch_to_device(iterator, size, devices=None):
   if devices is None:
     devices = jax.local_devices()
   def _prefetch(xs):
-    if hasattr(jax.api, "device_put_sharded"):  # jax>=0.2.0
-      return jax.api.device_put_sharded(list(xs), devices)
+    if hasattr(jax, "device_put_sharded"):  # jax>=0.2.0
+      return jax.device_put_sharded(list(xs), devices)
     else:
       aval = jax.xla.abstractify(xs)
       assert xs.shape[0] == len(devices), (
