@@ -27,11 +27,18 @@ from typing import Any, Type, Callable, Union, Mapping, Optional, TypeVar, Itera
 import dataclasses
 import functools
 import inspect
+from flax import errors
+from flax import traceback_util
 from flax.core import lift, Scope
 from flax.linen.module import Module
+from flax.linen.module import Variable
 from flax.linen.module import wrap_method_once
+from flax.linen import module as linen_module
+from flax import struct
 import jax
 
+traceback_util.register_exclusion(__file__)
+
 # Utils
 # -----------------------------------------------------------------------------
 def clean_clone(x):
@@ -44,59 +51,135 @@ def clean_clone(x):
   return x
 
 
-def get_module_scopes(module):
+@struct.dataclass
+class VariablePlaceholder:
+  """Used to mark Variables in a JAX-compatible way when lifting arguments."""
+  collection: str = struct.field(pytree_node=False)
+  name: str = struct.field(pytree_node=False)
+
+
+@struct.dataclass
+class InstancePlaceholder:
+  """Marks module instances in a JAX-compatible way when lifting arguments."""
+  cls: type = struct.field(pytree_node=False)
+  attrs: dict = struct.field(pytree_node=False)
+
+
+def get_module_scopes(module, args=None, kwargs=None):
   """Get all scopes on module, including constructor Module arguments.
 
   To properly functionalize a Module that has other bound Modules passed in
   "from the outside" as dataclass attributes, we need to traverse all dataclass
   fields to find the Scopes associated with the Module.  Additionally, because
   we allow Modules to be passed inside pytrees on the dataclass attributes, we
-  must traverse all dataclass attributes as pytrees to find all Modules.
+  must traverse all dataclass attributes as pytrees to find all Modules.  We
+  additionally handle lifting Variables (which are just references to data in
+  particular scopes) and Module instances that are passed as arguments to
+  methods.
 
   Args:
     module: a bound flax Module.
+    args: an *args list possibly containing Variables or Module instances
+      referencing a scope.
+    kwargs: a **kwargs dict possibly containing Variables or Module instances
+      referencing a scope.
 
   Returns:
     A list of all functional-core Scopes bound on self and inside dataclass
-    fields.
+    fields as well as any Scopes passed via argument Variables, an updated args
+    list, and an updated kwargs dict that have both had Variables replaced with
+    VariablePlaceholders and Module instances replaced with InstancePlaceholders
+    that are compatible with jax functions.
   """
   module._try_setup(shallow=True)
   outer_scopes = []
+  # gather scopes associated with Variables and Module instances passed as arguments
+  def get_arg_scope(x):
+    nonlocal outer_scopes
+    if isinstance(x, Variable) and isinstance(x.scope, Scope):
+      outer_scopes.append(x.scope)
+      return VariablePlaceholder(x.collection, x.name)
+    elif isinstance(x, Module) and isinstance(x.scope, Scope):
+      x._try_setup(shallow=True)
+      outer_scopes.append(x.scope)
+      attrs = {f.name: getattr(x, f.name)
+          for f in dataclasses.fields(x) if f.name != 'parent' and f.init}
+      attrs = jax.tree_map(get_arg_scope, attrs)
+      return InstancePlaceholder(x.__class__, attrs)
+    return x
+  new_args, new_kwargs = jax.tree_map(get_arg_scope, (args, kwargs))
+  # gather scopes in Variables and Submodules passed as Module attributes
   def get_scope(x):
     nonlocal outer_scopes
     if isinstance(x, Module) and isinstance(x.scope, Scope):
-      outer_scopes.extend(get_module_scopes(x))
+      module_scopes, _, _ = get_module_scopes(x)
+      outer_scopes.extend(module_scopes)
+    elif isinstance(x, Variable) and isinstance(x.scope, Scope):
+      outer_scopes.append(x.scope)
     return x
   attrs = {f.name: getattr(module, f.name)
            for f in dataclasses.fields(module) if f.name != 'parent' and f.init}
   jax.tree_map(get_scope, attrs)
-  return outer_scopes + [module.scope,]
+  return outer_scopes + [module.scope,], new_args, new_kwargs
 
 
-def set_module_scopes(module, scopes):
+def set_module_scopes(module, args, kwargs, scopes):
   """Set all scopes on module, including those on Modules in dataclass fields.
 
   To properly functionalize a Module we must also "rehydrate" it with Scopes
   from `get_module_scopes`.  We need to set scopes not just on the Module but
   also on any Module living inside dataclass attributes or even pytrees in its
-  dataclass attributes.  The order of traversal through both methods is the
-  same, guaranteeing the correct Scopes are applied to each Module.
+  dataclass attributes.  We additionally handle restoring Variables and Module
+  instances from their placeholders in the method positional and keyword
+  arguments.  The order of traversal through this method is the same as in
+  `get_module_scopes`, guaranteeing the correct Scopes are applied to each
+  Module.
 
   Args:
     module: a flax Module.
+    args: an *args list possibly containing VariablePlaceholder or
+      InstancePlaceholder members.
+    kwargs: a **kwargs dict possibly containing VariablePlaceholder or
+      InstancePlaceholder members.
     scopes: a list of Scopes corresponding to this Module and its arguments that
       was created by the `get_module_scopes` function.
 
   Returns:
     A copy of the module with it and its attributes bound to the scopes passed
-    to this function.
+    to this function, an updated args list, and an updated kwargs dict with
+    updated Variable and Module instance references.
   """
   idx = 0
+  # set scopes associated with Variables and Module instances passed as arguments
+  def set_arg_scope(x):
+    nonlocal idx
+    if isinstance(x, VariablePlaceholder):
+      new_x = Variable(scope=scopes[idx], collection=x.collection, name=x.name)
+      idx += 1
+      return new_x
+    elif isinstance(x, InstancePlaceholder):
+      instance_scope = scopes[idx]
+      idx += 1
+      instance_attrs = jax.tree_map(set_arg_scope, x.attrs)
+      return x.cls(parent=instance_scope, **instance_attrs)
+    return x
+  is_placeholder = lambda x: isinstance(x, (VariablePlaceholder, InstancePlaceholder))
+  new_args, new_kwargs = jax.tree_map(set_arg_scope,
+                                      (args, kwargs),
+                                      is_leaf=is_placeholder)
+  # set scopes in Variables and Submodules passed as Module attributes
   def set_scopes(module):
     nonlocal idx
     def set_scopes_inner(x):
+      nonlocal idx
       if isinstance(x, Module) and isinstance(x.scope, Scope):
         return set_scopes(x)
+      elif isinstance(x, Variable) and isinstance(x.scope, Scope):
+        new_x = Variable(scope=scopes[idx],
+                         collection=x.collection,
+                         name=x.name)
+        idx += 1
+        return new_x
       else:
         return x
     attrs = {f.name: getattr(module, f.name)
@@ -107,7 +190,15 @@ def set_module_scopes(module, scopes):
     return new_module
   new_module = set_scopes(module)
   assert len(scopes) == idx, f'scope list mismatch {len(scopes)} != {idx}'
-  return new_module
+  return new_module, new_args, new_kwargs
+
+
+def _test_transformed_return_values(tree, method_name):
+  """Tests whether the return value contains any Modules or Variables."""
+  impure = any(jax.tree_leaves(
+      jax.tree_map(lambda x: isinstance(x, (Module, Variable)), tree)))
+  if impure:
+    raise errors.TransformedMethodReturnValueError(method_name)
 
 
 # Class lifting
@@ -134,6 +225,14 @@ def module_class_lift_transform(
         all args must be passed via methods kwarg.""")
     class_trafo_args = {k: ((), v) for k, v in methods.items()}
 
+  # Handle partially initialized module class constructors.
+  if (isinstance(module_class, functools.partial) and
+      issubclass(module_class.func, Module)):
+    partial_object = module_class
+    module_class = module_class.func
+  else:
+    partial_object = None
+
   def create_trans_fn(fn_name, fn_trafo_args):
     # get existing unbound method from class
     fn = getattr(module_class, fn_name)
@@ -148,22 +247,30 @@ def module_class_lift_transform(
                  for f in dataclasses.fields(self) if f.name != 'parent' and f.init}
         # we reference module_class, not self.__class__ to avoid infinite loop
         cloned = module_class(parent=None, **attrs)
-        cloned = set_module_scopes(cloned, scopes)
+        cloned, args, kwargs = set_module_scopes(cloned, args, kwargs, scopes)
         object.__setattr__(cloned, '_state', self._state.export())  # pylint: disable=protected-access
         res = fn(cloned, *args, **kwargs)
         self._state.reimport(cloned._state)  # pylint: disable=protected-access
+        _test_transformed_return_values(res, fn_name)
         return res
       # here we apply the given lifting transform to the scope-ingesting fn
       trafo_fn = transform(core_fn, *trafo_args, **trafo_kwargs)
-      ret = trafo_fn(get_module_scopes(self), *args, **kwargs)
+      module_scopes, args, kwargs = get_module_scopes(self, args, kwargs)
+      ret = trafo_fn(module_scopes, *args, **kwargs)
       return ret
     return wrapped_fn
   transformed_fns = {fn_name: create_trans_fn(fn_name, fn_trafo_args)
                      for fn_name, fn_trafo_args in class_trafo_args.items()}
   # construct new dynamic class w. transformed methods
-  transformed_cls = type(transform.__name__.capitalize() + module_class.__name__,
-                         (module_class,),
-                         transformed_fns)
+  transformed_cls = type(
+      transform.__name__.capitalize() + module_class.__name__,
+      (module_class,),
+      transformed_fns)
+  # Handle partially initialized module class constructors.
+  if partial_object is not None:
+    transformed_cls = functools.partial(transformed_cls,
+                                        *partial_object.args,
+                                        **partial_object.keywords)
   return transformed_cls
 
 
@@ -177,14 +284,16 @@ def decorator_lift_transform(transform, class_fn, *trafo_args, **trafo_kwargs):
   def wrapped_fn(self, *args, **kwargs):
     # make a scope-function to transform
     def core_fn(scopes, *args, **kwargs):
-      cloned = set_module_scopes(self, scopes)
+      cloned, args, kwargs = set_module_scopes(self, args, kwargs, scopes)
       object.__setattr__(cloned, '_state', self._state.export())  # pylint: disable=protected-access
       res = prewrapped_fn(cloned, *args, **kwargs)
       self._state.reimport(cloned._state)  # pylint: disable=protected-access
+      _test_transformed_return_values(res, class_fn.__name__)
       return res
     # here we apply the given lifting transform to the scope-ingesting fn
     trafo_fn = transform(core_fn, *trafo_args, **trafo_kwargs)
-    return trafo_fn(get_module_scopes(self), *args, **kwargs)
+    module_scopes, args, kwargs = get_module_scopes(self, args, kwargs)
+    return trafo_fn(module_scopes, *args, **kwargs)
   return wrapped_fn
 
 
@@ -192,7 +301,8 @@ def decorator_lift_transform(transform, class_fn, *trafo_args, **trafo_kwargs):
 # -----------------------------------------------------------------------------
 def lift_transform(transform, target, *trafo_args, methods=None, **trafo_kwargs):
   """Applies to class or as a decorator on class fns."""
-  if inspect.isclass(target) and issubclass(target, Module):
+  if (inspect.isclass(target) and issubclass(target, Module)
+      or isinstance(target, functools.partial)):
     return module_class_lift_transform(
         transform, target, *trafo_args, methods=methods, **trafo_kwargs)
   # we presume this is being used as a function decorator in class definition
@@ -209,6 +319,7 @@ TransformTarget = Union[Type[Module], Callable[..., Any]]
 
 Target = TypeVar('Target', bound=TransformTarget)
 
+
 def vmap(target: Target,
          variable_axes: Mapping[lift.CollectionFilter, lift.InOutAxis],
          split_rngs: Mapping[lift.PRNGSequenceFilter, bool],
@@ -223,7 +334,7 @@ def vmap(target: Target,
   ``vmap`` can be used to add a batch axis to a ``Module``.
   For example we could create a version of ``Dense`` with
   a batch axis that does not share parameters::
-  
+
     BatchDense = nn.vmap(
         nn.Dense,
         in_axes=0, out_axes=0,
@@ -281,13 +392,13 @@ def jit(target: Target,
         backend: Union[str, None] = None,
         methods=None) -> Target:
   """Lifted version of ``jax.jit``.
-  
+
   Args:
     target: a ``Module`` or a function taking a ``Module``
       as its first argument.
     variables: The variable collections that are lifted. By default all
       collections are lifted.
-    rngs: The PRNG sequences that are lifted. By defualt all PRNG sequences
+    rngs: The PRNG sequences that are lifted. By default all PRNG sequences
       are lifted.
     static_argnums: An int or collection of ints specifying which positional
       arguments to treat as static (compile-time constant). Operations that only
@@ -331,9 +442,10 @@ def checkpoint(target: Target,
         variables: lift.CollectionFilter = True,
         rngs: lift.PRNGSequenceFilter = True,
         concrete: bool = False,
+        prevent_cse: bool = True,
         methods=None) -> Target:
   """Lifted version of ``jax.checkpoint``.
-  
+
   This function is aliased to ``lift.remat`` just like ``jax.remat``.
 
   Args:
@@ -342,21 +454,29 @@ def checkpoint(target: Target,
       re-computed when computing gradients for the target.
     variables: The variable collections that are lifted. By default all
       collections are lifted.
-    rngs: The PRNG sequences that are lifted. By defualt all PRNG sequences
+    rngs: The PRNG sequences that are lifted. By default all PRNG sequences
       are lifted.
     concrete: Optional, boolean indicating whether ``fun`` may involve
       value-dependent Python control flow (default False). Support for such
       control flow is optional, and disabled by default, because in some
       edge-case compositions with :func:`jax.jit` it can lead to some extra
       computation.
+    prevent_cse: Optional, boolean indicating whether to prevent common
+      subexpression elimination (CSE) optimizations in the HLO generated from
+      differentiation. This CSE prevention has costs because it can foil other
+      optimizations, and because it can incur high overheads on some backends,
+      especially GPU. The default is True because otherwise, under a ``jit`` or
+      ``pmap``, CSE can defeat the purpose of this decorator. But in some
+      settings, like when used inside a ``scan``, this CSE prevention mechanism
+      is unnecessary, in which case ``prevent_cse`` should be set to False.
   Returns:
     A wrapped version of ``target``. When computing gradients intermediate
     computations will be re-computed on the backward pass.
   """
   return lift_transform(
       lift.checkpoint, target,
-      variables=variables, rngs=rngs,
-      concrete=concrete, methods=methods)
+      variables=variables, rngs=rngs, concrete=concrete,
+      prevent_cse=prevent_cse, methods=methods)
 
 
 remat = checkpoint
@@ -370,6 +490,7 @@ def scan(target: Target,
          in_axes=0, out_axes=0,
          length: Optional[int] = None,
          reverse: bool = False,
+         data_transform: Optional[Callable[..., Any]] = None,
          methods=None) -> Target:
   """A lifted version of ``jax.lax.scan``.
 
@@ -396,20 +517,31 @@ def scan(target: Target,
 
   Example::
 
+    import flax
+    import flax.linen as nn
+    from jax import random
+
     class SimpleScan(nn.Module):
       @nn.compact
       def __call__(self, c, xs):
         LSTM = nn.scan(nn.LSTMCell,
                        variable_broadcast="params",
-                       split_rngs={"params": False})
+                       split_rngs={"params": False},
+                       in_axes=1,
+                       out_axes=1)
         return LSTM()(c, xs)
 
-    xs = random.uniform(rng_1, (batch_size, features))
-    carry_0 = nn.LSTMCell.initialize_carry(
-        random.PRNGKey(0), (batch_size,), features)
+    seq_len, batch_size, in_feat, out_feat = 20, 16, 3, 5
+    key_1, key_2, key_3 = random.split(random.PRNGKey(0), 3)
+
+    xs = random.uniform(key_1, (batch_size, seq_len, in_feat))
+    init_carry = nn.LSTMCell.initialize_carry(key_2, (batch_size,), out_feat)
+
     model = SimpleScan()
-    variables = model.init(key_2, carry_0, xs)
-    out_state, out_val = model.apply(variables, carry_0, xs)
+    variables = model.init(key_3, init_carry, xs)
+    out_carry, out_val = model.apply(variables, init_carry, xs)
+
+    assert out_val.shape == (batch_size, seq_len, out_feat)
 
 
   Args:
@@ -432,6 +564,10 @@ def scan(target: Target,
     length: Specifies the number of loop iterations. This only needs
       to be specified if it cannot be derivied from the scan arguments.
     reverse: If true, scan from end to start in reverse order.
+    data_transform: optional function to transform raw functional-core variable
+      and rng groups inside lifted scan body_fn, intended for inline SPMD
+      annotations.
+
   Returns:
     The scan function with the signature ``(scope, carry, *xxs) -> (carry, yys)``,
     where ``xxs`` and ``yys`` are the scan values that go in and out of the loop.
@@ -445,29 +581,39 @@ def scan(target: Target,
       in_axes=in_axes, out_axes=out_axes,
       length=length,
       reverse=reverse,
+      data_transform=data_transform,
       methods=methods)
 
 
 # Special case of decorator_lift_transform to handle named calls for profiling.
-def named_call(class_fn):
-  """Labels a method for labelled traces in profiles."""
+def named_call(class_fn, force=True):
+  """Labels a method for labelled traces in profiles.
+  
+  Args:
+    force: If True, the named_call transform is applied even if it is globally disabled.
+      (e.g.: by calling `flax.linen.disable_named_call()`)
+  """
   # Due to the ordering of method decorators, we must wrap the class_fn
   # with the module state management wrapper first to maintain Module state correctly.
   prewrapped_fn = wrap_method_once(class_fn)
   @functools.wraps(prewrapped_fn)
   def wrapped_fn(self, *args, **kwargs):
+    if not force and not linen_module._use_named_call:
+      return prewrapped_fn(self, *args, **kwargs)
     fn_name = class_fn.__name__
     method_suffix = f'.{fn_name}' if fn_name != '__call__' else ''
     module_name = self.name or self.__class__.__name__
     full_name = f'{module_name}{method_suffix}'
     # make a scope-function to transform
     def core_fn(scopes, *args, **kwargs):
-      cloned = set_module_scopes(self, scopes)
+      cloned, args, kwargs = set_module_scopes(self, args, kwargs, scopes)
       object.__setattr__(cloned, '_state', self._state.export())  # pylint: disable=protected-access
       res = prewrapped_fn(cloned, *args, **kwargs)
       self._state.reimport(cloned._state)  # pylint: disable=protected-access
+      _test_transformed_return_values(res, fn_name)
       return res
     # here we apply the given lifting transform to the scope-ingesting fn
     trafo_fn = lift.named_call(core_fn, full_name)
-    return trafo_fn(get_module_scopes(self), *args, **kwargs)
+    module_scopes, args, kwargs = get_module_scopes(self, args, kwargs)
+    return trafo_fn(module_scopes, *args, **kwargs)
   return wrapped_fn
